from typing_extensions import dataclass_transform
import attrs
import frozendict
from uuid import uuid4
import json

from typing import Self, TypeVar


def convert_metric_field(value):
    match value:
        case dict():
            return frozendict.deepfreeze(value)
        case list():
            return tuple(value)
        case _:
            return value


def freeze_metric_fields(cls, fields):
    """
    Instances of Metric are intended to be serialised to JSON, and must be immutable.
    We'd like to have a __hash__ implementation, but dicts and lists are both mutable and unhashable.
    This recursively converts values to frozendicts (https://github.com/Marco-Sulla/python-frozendict) and tuples, respectively.
    """
    return [f.evolve(converter=convert_metric_field) for f in fields]


class RunScope:
    """
    Base marker type for dataclass-like structs containing JSON-isomorphic parameter data for a context level.

    This is an attrs frozen class, so subclasses should not rely on an __init__ method to be called.

    Defining __attrs_post_init__ should work though.
    """

    def asdict(self):
        return attrs.asdict(self)

    @classmethod
    def from_canonical_json(cls, json_str: str) -> Self:
        return cls(**json.loads(json_str))

    def as_canonical_json(self):
        # TODO: actually implement the "canonical" part
        return json.dumps(self.asdict())

    def namespace(self) -> str:
        raise NotImplementedError()

    def scope_name(self) -> str:
        return type(self).__name__


T = TypeVar("T", bound=type[RunScope])


def make_scope(cls: T) -> T:
    """Class decorator for registering class names and ensuring inheritance from Metric."""
    assert issubclass(cls, RunScope)

    cls = attrs.frozen(field_transformer=freeze_metric_fields)(cls)
    return cls



from permanence import ObjectPermanence, Arty, ScopeNamespace
from permanence.tag_type import Tagged
from permanence.chunks import (
    LocalChunk,
    MutableLocalChunk,
    ChunkCollection,
    TaggedLocalArtifact,
)
from permanence.artifact import TableArtifact

import pandas as pd

import asyncio
import time


def test_chunk_append():
    df = pd.DataFrame({"x": [1, 2, 3], "z": ["a", "b", "d"]})

    ch = MutableLocalChunk.new()
    af_bin, af = Arty.Table.Foo.new(df)
    ch.append(af_bin, TaggedLocalArtifact(df, af, Tagged({}), 0.0))

    chc = ChunkCollection([ch])
    assert not chc.query().empty


def test_chunk_roundtrip():
    space = ScopeNamespace("just-for-testing_chunks")
    pr = ObjectPermanence(space)

    @pr.new_scope
    class Params(pr.Scope):
        n: int

    df_a = pd.DataFrame({"x": [1, 2, 3], "z": ["a", "b", "d"]})
    df_b = pd.DataFrame({"q": [1, 2, 3], "m": ["a", "b", "d"]})

    with pr.context(Params(1001)):
        something_af = pr.logged(Arty.Table.Something, df_a)
        assert isinstance(something_af, TableArtifact)
        print(something_af.data_hash)

        assert pr.query(values=Arty.Table.Something).only().artifact == something_af

        with pr.using(something_af) as _:
            pr.logged(Arty.Table.SomethingElse, df_b)

    metric_log = {"metric:accuracy": 1000000000.0}

    with pr.context(Params(1002)):
        # Test all artifact types
        pr.logged(Arty.Table.Something, df_a)
        pr.logged(Arty.JSON.MetricsOrIDK, metric_log)

    new_doc_id = asyncio.run(pr.chunks().into_document())  # Upload

    pr_alt = ObjectPermanence(space)

    assert pr_alt.query().empty

    read_chunks = asyncio.run(
        ChunkCollection.document_to_local_chunks(pr_alt.scope_namespace, new_doc_id)
    )
    assert len(read_chunks.query()) == 4

    # Test chunk API and behaviour on loaded contents

    with pr_alt.overlayed(read_chunks):
        assert len(pr_alt.query()) == 4

        # Map hash -> artifact
        assert pr_alt.checkout(something_af.data_hash) == something_af

        with pr_alt.context(something_af):
            matched = pr_alt.query().only()
            assert matched.artifact.scope_name() == "SomethingElse"
            assert isinstance(matched.artifact, TableArtifact)
            assert (matched.load().to_pandas() == df_b).all().all()

        # Test log time was within 5 seconds ago
        assert (
            0
            < time.time() - pr_alt.query(values=Arty.JSON.MetricsOrIDK).only().logged_at
            < 5
        )

        assert pr_alt.query(values=Arty.JSON.MetricsOrIDK).only().load() == metric_log

        # Map artifact -> contents
        with pr_alt.using(something_af) as something:
            assert (something.to_pandas() == df_a).all().all()

    assert pr_alt.query().empty



from .scope_type import RunScope, make_scope

from typing_extensions import dataclass_transform


class ScopeNamespace:
    def __init__(self, named_identifier: str):
        self.name = named_identifier

        self.registered_artifacts = {}
        self.registered_scopes: dict[str, type[RunScope]] = {}

        # TODO: Sanitise scope name?
        self.scope_cls = type(
            self.name, (RunScope,), {"namespace": lambda _: self.name}
        )

    @property
    def Scope(self) -> type[RunScope]:
        return self.scope_cls

    @dataclass_transform()
    def new_scope(self, cls) -> type[RunScope]:
        new_cls = type(cls.__name__, (self.scope_cls,), {**cls.__dict__})
        new_cls = make_scope(new_cls)

        # TODO: do some sort of "class equality" check
        # assert new_cls.__name__ not in self.registered_scopes
        if new_cls.__name__ in self.registered_scopes:
            return self.registered_scopes[new_cls.__name__]
            # Error: only supported for mutable types or...
            # self.registered_scopes[new_cls.__name__].__class__ = new_cls

        self.registered_scopes[new_cls.__name__] = new_cls
        return new_cls

    def new_artifact(self, cls):
        return cls



from __future__ import annotations

import asyncio
from functools import reduce
from permanence.namespace import ScopeNamespace
from permanence.tag_type import Tagged
from .artifact import Artifact, Arty, JSONArtifact, TableArtifact
from .scope_type import RunScope

import iroh
from iroh import Iroh
import pyarrow as pa
import pyarrow.feather as ft
from uuid6 import uuid7

from io import BytesIO
import itertools
from dataclasses import dataclass, field
from collections import defaultdict
from typing import Iterator, Iterable, Any, Literal, Optional, Protocol, Sequence

from loguru import logger


"""
# Chunk data model

Remote

"""


class TaggedArtifact(Protocol):
    artifact: Artifact
    tag: Tagged
    # TODO: datetime maybe? or convert via method
    logged_at: float

    def load(self) -> Any: ...


@dataclass
class TaggedLocalArtifact(TaggedArtifact):
    artifact_content: Any
    artifact: Artifact
    tag: Tagged
    logged_at: float

    def load(self):
        return self.artifact_content


@dataclass
class TaggedRemoteArtifact(TaggedArtifact):
    artifact: Artifact
    # generate "mock" scopes via attrs instead of registry?
    tag: Tagged
    logged_at: float

    def load(self):
        raise NotImplementedError()


class Chronos:
    """
    Wrapper for a chronologically ordered iterator (usually derived from a collection of TaggedArtifact via log timestamp).
    """

    def __init__(self, iterator):
        self.iterator = iterator
        self.consumed = None

    def collected(self):
        if self.consumed is not None:
            return self.consumed
        else:
            self.consumed = list(self.iterator)
            return self.consumed

    def __iter__(self):
        yield from self.collected()

    def __len__(self):
        return len(self.collected())

    @property
    def empty(self):
        return len(self) == 0

    def earliest(self):
        return self.collected()[0]

    def latest(self):
        return self.collected()[-1]

    def only(self):
        assert len(self) == 1, self.collected()
        return self.latest()


@dataclass
class LocalChunk:
    """
    Blobs are stored fully deserialised in their respective Python types:
    - Table => pyarrow.Table
    - JSON => dict
    """

    # TODO: remove default initialisers (in `new` instead)

    chunk_id: str
    "UUIDv7 identifying the chunk"

    reverse_artifacts: dict[RunScope, set[str]] = field(
        default_factory=lambda: defaultdict(lambda: set())
    )
    "Maps scopes to the set of artifact hashes logged within them"

    art_hashes: dict[str, list[Artifact]] = field(
        default_factory=lambda: defaultdict(lambda: [])
    )
    """
    Maps content hashes to matching artifacts.
    
    Although hashes uniquely (mod collisions) uniquely identify content, they're mapped to a list because `Artifact` equality also requires types to match.
    """

    art_logs: dict[Artifact, list[TaggedLocalArtifact]] = field(
        default_factory=lambda: defaultdict(lambda: [])
    )
    """
    Maps artifacts to the contexts they've been logged in.

    Values are lists because despite containing identical content (as references into art_storage) 
    the same blob can be logged multiple times under different tags.
    """

    blob_storage: dict[str, bytes] = field(default_factory=lambda: {})
    "Hash to binary blob lookup. To eventually be replaced with a global blob store/cache of calls to Iroh storage"

    @classmethod
    def new(cls):
        return cls(chunk_id=str(uuid7()))

    def copy(self):
        # Always returns LocalChunk (strips mutability)
        return LocalChunk(
            chunk_id=self.chunk_id,
            reverse_artifacts=self.reverse_artifacts.copy(),
            art_hashes=self.art_hashes.copy(),
            art_logs=self.art_logs.copy(),
            blob_storage=self.blob_storage.copy(),
        )

    def filter_scope_intersection(self, scopes: Iterable[RunScope]) -> set[Artifact]:
        """
        Finds artifacts that have been logged at least once in each of the given scopes.

        Warning! This might not do what you think it does. Because this operates on artifacts rather than instances, if scopes is `[A(), B()]` this will return artifacts logged once under A() and once separately under B(), in addition to those logged under `[A(), B()]` jointly.
        """
        scopes = list(scopes)
        if not scopes:
            return set().union(*self.art_hashes.values())

        artifact_hashes = reduce(
            set.intersection, map(self.reverse_artifacts.__getitem__, scopes)
        )
        return reduce(
            set.union, map(self.art_hashes.__getitem__, artifact_hashes), set()
        )

    def match_scope_intersection(self, match_scopes) -> list[TaggedLocalArtifact]:
        match_scopes = set(match_scopes)
        candidate_artifacts = self.filter_scope_intersection(match_scopes)
        logs_stream = itertools.chain.from_iterable(
            map(self.art_logs.__getitem__, candidate_artifacts)
        )
        matched = filter(lambda taf: taf.tag.matches_query(match_scopes), logs_stream)
        return list(matched)

    def serialise(self) -> dict[str, bytes]:
        """
        Converts into Iroh document {keys => binary blob} format.

        This does not include artifact contents which need to be sent to a content-addressed-storage backend like Iroh that can map BLAKE3 hashes to arbitrary binary content.
        """
        # TODO: Use package __version__
        metadata = {"format_version": "0.0.1"}

        # Serialise scopes
        scopes = {}

        def scope_with_hash(scope: RunScope, scope_id: str, saw_obj_hash: str):
            nonlocal scopes

            if scope_id not in scopes:
                scopes[scope_id] = {
                    "name": type(scope).__name__,
                    "content": scope.as_canonical_json(),
                    "namespace": scope.namespace(),
                    "objects": [],
                }
            scopes[scope_id]["objects"].append(saw_obj_hash)
            return scope_id

        # Serialise logs
        rows = []

        iter_artifacts = itertools.chain.from_iterable(self.art_logs.values())
        for tagged_artifact in iter_artifacts:
            artifact = tagged_artifact.artifact
            tagged = tagged_artifact.tag
            row_ser = {
                "content_binary_local": None,
                "content_binary_blake3": artifact.data_hash,
                "log_metadata": {"logged_at": tagged_artifact.logged_at},
                "artifact_name": artifact.scope_name(),
                "artifact_type": artifact.type_string(),
                "context_ids": [
                    scope_with_hash(s, h, artifact.data_hash)
                    for s, h in tagged.scope_state.items()
                ],
                "contexts_hash": tagged.hashed(),
            }

            # For small-ish values, storing inline is way faster
            if isinstance(artifact, JSONArtifact):
                row_ser["content_binary_local"] = artifact.new(
                    tagged_artifact.artifact_content
                )[0]
            rows.append(row_ser)

        # also specify reciprocal hashes to ensure batches can be matched together?
        scope_table = pa.Table.from_pylist(
            [{"context_id": k, **v} for k, v in scopes.items()], metadata=metadata
        )
        log_table = pa.Table.from_pylist(rows, metadata=metadata)

        # TODO: declare Arrow schema - to handle case of empty tables
        write_logs = BytesIO()
        with pa.ipc.new_file(write_logs, log_table.schema) as writer:
            writer.write_table(log_table)

        write_scopes = BytesIO()
        with pa.ipc.new_file(write_scopes, scope_table.schema) as writer:
            writer.write_table(scope_table)

        return {
            # TODO: more chunk/scope/session metadata?
            "contexts": write_scopes.getvalue(),
            "logs": write_logs.getvalue(),
        }

    @staticmethod
    def deserialise(
        scope_namespace: ScopeNamespace,
        doc_keys: dict[str, bytes],
        doc_blobs: dict[str, bytes],
        chunk_id: str,
    ):
        def reconstruct_scope_type(row: dict):
            if row["namespace"].startswith("."):
                # Artifact scopes get special path
                match row["namespace"].split("."):
                    case ["", "Artifact", artifact_cls]:
                        artifact_type = getattr(
                            getattr(Arty, artifact_cls), row["name"]
                        )
                        return artifact_type
                    case _:
                        raise ValueError(f"{row}: expected artifact namespace")
            else:
                return scope_namespace.registered_scopes[row["name"]]

        # TODO: return Result
        log_key = f"{chunk_id}/logs"
        scope_key = f"{chunk_id}/contexts"
        assert log_key in doc_keys
        assert scope_key in doc_keys

        log_table = ft.read_table(BytesIO(doc_keys[log_key]))
        scope_table = ft.read_table(BytesIO(doc_keys[scope_key]))

        contexts = scope_table.to_pylist()
        # TODO: reconstruct all contexts in one pass here?
        contexts = {c["context_id"]: c for c in contexts}

        builder = MutableLocalChunk.new()

        for log in log_table.to_pylist():
            match log["artifact_type"]:
                case "Table":
                    artifact_type = getattr(Arty.Table, log["artifact_name"])
                    artifact = artifact_type(
                        log["content_binary_blake3"],
                    )
                    content_blob = doc_blobs[log["content_binary_blake3"]]
                    artifact_content = artifact_type.reconstruct(content_blob)
                case "JSON":
                    artifact_type = getattr(Arty.JSON, log["artifact_name"])
                    artifact = artifact_type(
                        log["content_binary_blake3"],
                    )
                    content_blob = log["content_binary_local"]
                    artifact_content = artifact_type.reconstruct(content_blob)
                case unknown_artifact:
                    raise TypeError(f"{unknown_artifact}: unknown artifact type")

            tag = Tagged(
                {
                    reconstruct_scope_type(contexts[ctx_id]).from_canonical_json(
                        contexts[ctx_id]["content"]
                    ): ctx_id
                    for ctx_id in log["context_ids"]
                }
            )
            # TODO: load log timestamp
            tla = TaggedLocalArtifact(
                artifact_content,
                artifact,
                tag,
                log["log_metadata"].get("logged_at", 0.0),
            )
            builder.append(content_blob, tla)

        return builder.copy()


class MutableLocalChunk(LocalChunk):
    def append(self, af_bin: bytes, tla: TaggedLocalArtifact):
        af = tla.artifact

        for scope in tla.tag:
            self.reverse_artifacts[scope].add(af.data_hash)

        self.art_hashes[af.data_hash].append(af)

        self.art_logs[af].append(tla)

        self.blob_storage[af.data_hash] = af_bin


class RemoteChunk:
    """
    Keeps a handle to the Iroh worker/sync task and will fetch on-demand.
    """

    pass


class ChunkQueryMixin:
    """
    The result of a query associates each chunk with a subset of matching artifacts.
    """

    def iter_chunks(self) -> Iterator[LocalChunk | RemoteChunk]:
        raise NotImplementedError()

    def query(
        self,
        scopes: Sequence[RunScope] = (),
        values: Sequence[type[Artifact]] | type[Artifact] | None = None,
    ):
        # TODO: support remote chunks
        return ChunkQuerySet.from_query(self.iter_chunks(), scopes, values)

    def instances(self, af: Artifact) -> Iterator[TaggedLocalArtifact]:
        raise NotImplementedError()


class ChunkCollection(ChunkQueryMixin):
    """
    Queries defined and executed on a set of chunks

    We can "load" a document, converting all remote chunks into local ones
    """

    def __init__(self, chunks: list[LocalChunk]):
        # TODO: common API for remote chunks
        # TODO: this should really be a map chunk_id => chunk
        self._chunks = list(chunks)

    def iter_chunks(self):
        yield from self._chunks

    def fetch_instances_by_hash(self, b3sum: str) -> Iterator[TaggedLocalArtifact]:
        # TODO: handle remote chunks
        for chunk in self.iter_chunks():
            for artifact in chunk.art_hashes[b3sum]:
                yield from self.fetch_instances(artifact)

    def fetch_instances(self, af: Artifact) -> Iterator[TaggedLocalArtifact]:
        # TODO: handle remote chunks
        for chunk in self.iter_chunks():
            yield from chunk.art_logs[af]

    @classmethod
    async def persistent_store(cls):
        # TODO: expose path configuration
        return await Iroh.persistent(".iroh-perm-store")

    @classmethod
    async def document_to_local_chunks(
        cls, scope_namespace: ScopeNamespace, document_id: str
    ):
        node: iroh.Iroh = await cls.persistent_store()
        doc = await node.docs().open(document_id)
        assert doc is not None

        entries = await doc.get_many(iroh.Query.all(None))

        # TODO: Batch API? Parallel async?
        blob_keys = await node.blobs().list()
        blobs = {k.to_hex(): await node.blobs().read_to_bytes(k) for k in blob_keys}

        doc_dict = {}
        for entry in entries:
            key = entry.key().decode()
            doc_dict[key] = await entry.content_bytes(doc)

        found_chunk_ids = set(k.split("/")[0] for k in doc_dict.keys())
        chunks = []
        for chunk_id in found_chunk_ids:
            chunks.append(
                LocalChunk.deserialise(scope_namespace, doc_dict, blobs, chunk_id)
            )
        return ChunkCollection(chunks)

    @staticmethod
    async def from_shared_document(space, node: Iroh, ticket):
        doc = await node.docs().join(ticket)
        raise NotImplementedError()

    async def into_document(self, doc_id: Optional[str] = None) -> str:
        node: iroh.Iroh = await self.persistent_store()
        if doc_id is None:
            doc: iroh.Doc = await node.docs().create()
        else:
            doc = await node.docs().open(doc_id)  # type: ignore
            assert doc is not None, "Document does not exist"

        logger.info(f"Created document {doc.id()}")
        whoami = (
            await node.authors().create()
        )  # perhaps try load from environment variable?

        for chunk in self.iter_chunks():
            if isinstance(chunk, RemoteChunk):
                # Presumably we'd just need to set the *hashes* without having to upload blobs, assuming it's the same node?
                raise NotImplementedError()

            for key, data in chunk.serialise().items():
                await doc.set_bytes(whoami, f"{chunk.chunk_id}/{key}".encode(), data)

            for _blob_hash, blob_bin in chunk.blob_storage.items():
                await node.blobs().add_bytes(blob_bin)

        await asyncio.sleep(1)
        await node.node().shutdown(False)
        await asyncio.sleep(1)  # Cleanup?

        return doc.id()

    def __add__(self, other: ChunkCollection):
        return ChunkCollection([*self.iter_chunks(), *other.iter_chunks()])


class ChunkQuerySet(Chronos):
    # TODO: Implement ChunkQueryMixin for this as well!
    def __init__(self, matches: dict[type[Artifact], list[TaggedLocalArtifact]]):
        self.matches = matches
        super().__init__(self.collect_ordered())

    def collect_ordered(self) -> list[TaggedLocalArtifact]:
        x = []
        for v in self.matches.values():
            x.extend(v)
        x.sort(key=lambda tla: tla.logged_at)
        return x

    def iter_artifacts(self):
        # Note: ordered by time *first* encountered, so won't necessarily give you the thing you just logged
        seen_artifacts = set()
        for tla in self.collect_ordered():
            if tla.artifact not in seen_artifacts:
                seen_artifacts.add(tla.artifact)
                yield tla.artifact

    def artifacts(self):
        return Chronos(self.iter_artifacts())

    def values(self):
        return Chronos(map(lambda tla: tla.artifact_content, self.collect_ordered()))

    @staticmethod
    def from_query(
        iter_chunks: Iterator[LocalChunk],
        scopes: Sequence[RunScope],
        values: Sequence[type[Artifact]] | type[Artifact] | None,
    ):
        """
        Scope queries behave like intersections (ALL scopes must match).
        Value queries are unions (artifact type matching ANY one in the query set).
        """
        # TODO: Support RemoteChunk
        match values:
            case None:
                value_queries = None
            case type():
                value_queries = {values}
            case _:
                value_queries = set(values)

        def require_artifact_of(af: Artifact):
            if value_queries is not None:
                return type(af) in value_queries
            else:
                return True

        by_type = defaultdict(list)
        for ch in iter_chunks:
            for la in ch.match_scope_intersection(scopes):
                if require_artifact_of(la.artifact):
                    by_type[type(la.artifact)].append(la)
        return ChunkQuerySet(by_type)

    def query(
        self,
        *,
        scopes: Sequence[RunScope] = (),
        values: Sequence[type[Artifact]] | type[Artifact] | None = None,
    ):
        raise NotImplementedError()

    def metrics_view(self, expand: Literal["field", "scope", "unit"] = "field"):
        """
        Converts JSON artifacts into a Pandas DataFrame, with scopes set as index columns.

        Expansion argument controls how scopes are turned into columns:
        - ...
        """

        import pandas as pd

        match expand:
            case "field":

                def expand_row(row):
                    res = {}
                    for scope in row["pr.scopes"]:
                        res.update(
                            {
                                f"{scope.scope_name()}.{k}": v
                                for k, v in scope.asdict().items()
                            }
                        )
                    return res
            case "scope":

                def expand_row(row):
                    res = {}
                    for scope in row["pr.scopes"]:
                        res[scope.scope_name()] = scope
                    return res

            case "unit":
                pass

            case _:
                raise ValueError(
                    f'{expand} must be a Literal["field", "scope", "unit"]'
                )

        def expand_df(df):
            if expand == "unit":
                return df.set_index("pr.scopes")
            else:
                non_index_cols = set(df.columns) - {"pr.scopes"}
                result = df.join(
                    df.apply(
                        expand_row,
                        result_type="expand",
                        axis=1,
                    )
                )
                del result["pr.scopes"]
                return result.set_index(
                    [c for c in result.columns if c not in non_index_cols]
                )

        rows_by_type = defaultdict(lambda: [])
        for tla in self:
            tla: TaggedLocalArtifact
            # Should loading happen at a chunk level?
            # TODO: in principle we could concatenate table artifacts as well
            if isinstance(tla.artifact, JSONArtifact):
                rows_by_type[tla.artifact.scope_name()].append(new_row := tla.load())
                new_row["pr.scopes"] = tla.tag
        return {af: expand_df(pd.DataFrame(rows)) for af, rows in rows_by_type.items()}



from permanence import ObjectPermanence
from attrs import asdict

import pytest


def test_disallow_record_subclassing():
    pr = ObjectPermanence.new_namespace("just-for-testing_nonsense")

    @pr.new_scope
    class Level(pr.Scope):
        name: str

    # subclassing is not allowed
    with pytest.raises(Exception):

        @pr.new_scope
        class LevelTwo(Level):
            name: str

    # forbid inconsistent scopes
    with pytest.raises(RuntimeError):
        with pr.context(Level("one")):
            with pr.context(Level("two")):
                pass


def test_different_namespaces():
    pr_A = ObjectPermanence.new_namespace("just-for-testing_nonsense_A")
    pr_b = ObjectPermanence.new_namespace("just-for-testing_nonsense_B")

    # At the module level you'd just define `Scope = A` and `new_scope = per_a.new_scope` and import those

    @pr_A.new_scope
    class Ope(pr_A.Scope):
        x: int

    ope_a = Ope

    @pr_b.new_scope
    class Ope(pr_b.Scope):
        x: int

    ope_b = Ope

    # equality checks must take namespace into account
    assert asdict(ope_a(1)) == asdict(ope_b(1))
    assert ope_a(1) != ope_b(1)
    assert ope_a(1).namespace() != ope_b(1).namespace()



from permanence import (
    ObjectPermanence,
    Scope,
    Arty,
)
import pytest


def test_init_and_log():
    pr = ObjectPermanence.new_namespace("just-for-testing_basic")

    @pr.new_scope
    class Basically(pr.Scope):
        name: str

    @pr.new_scope
    class Totally(pr.Scope):
        n: int

    with pr.context(Basically("bob")):
        pr.logged(Arty.JSON.XYZ, {"xyz": 256})

    with pr.context(Basically("eve")):
        pr.log({"abc": 1024})

    all_values = pr.query()
    assert len(all_values) == 2

    with pr.context(Basically("alice")):
        assert pr.query().empty

    with pr.context(Basically("bob")):
        assert pr.query().only().load() == {"xyz": 256}

    with pr.context(Totally(0)):
        pr.log({"id": 0})

    with pr.context(Basically("bob")):
        with pr.context(Totally(0)):
            assert pr.query().empty

    with pr.context(Totally(37)):
        assert pr[Totally].n == 37

    with pytest.raises(RuntimeError):
        # Not in context
        print(pr[Totally].n)


def test_scope_usage():
    # Uses default "global" script scope
    from permanence import pr

    @pr.new_scope
    class Ope(Scope):
        x: int

    with pr.context(Ope(22)):
        pr.log({"result": 1})



from .namespace import ScopeNamespace
from .scope_type import RunScope, make_scope
from .object import ObjectPermanence
from .artifact import Arty
from .chunks import ChunkCollection

pr = ObjectPermanence.new_namespace("main_program_scope")
Scope = pr.Scope
new_scope = pr.new_scope



import attrs
import frozendict

from typing import Self, TypeVar


def convert_metric_field(value):
    match value:
        case dict():
            return frozendict.deepfreeze(value)
        case list():
            return tuple(value)
        case _:
            return value


def freeze_metric_fields(cls, fields):
    """
    Instances of Metric are intended to be serialised to JSON, and must be immutable.
    We'd like to have a __hash__ implementation, but dicts and lists are both mutable and unhashable.
    This recursively converts values to frozendicts (https://github.com/Marco-Sulla/python-frozendict) and tuples, respectively.
    """
    return [f.evolve(converter=convert_metric_field) for f in fields]


class RunScope:
    """
    Base marker type for dataclass-like structs containing JSON-isomorphic parameter data for a context level.
    """

    @classmethod
    def get(cls) -> Self:
        # This method gets patched in later
        raise NotImplementedError()

    def asdict(self):
        return attrs.asdict(self)


T = TypeVar("T", bound=type[RunScope])


def make_scope(cls: T) -> T:
    """Class decorator for registering class names and ensuring inheritance from Metric."""
    assert issubclass(cls, RunScope)

    # disallow subclassing of other records
    scope_cls = [
        parent_cls
        for parent_cls in cls.mro()
        if issubclass(parent_cls, RunScope)
        and parent_cls.__module__.split(".")[0] != "permanence"
    ]
    assert len(scope_cls) == 1, "Subclassing a Record type is disallowed"

    cls = attrs.frozen(field_transformer=freeze_metric_fields)(cls)
    return cls


def new_scope_namespace(name: str):
    """
    Defines a unique namespace that will be considered completely independent of other namespaces (w.r.t scopes, objects and all other functions).

    Also helps with type-safety.
    """

    class NamedScope(RunScope):
        _scope_namespace = name

    return NamedScope



from .bounds import btest_rel

from permanence import RunScope
from permanence.chunks import ChunkQuerySet

import pandas as pd
import numpy as np

from dataclasses import dataclass
from typing import Any, Literal

from loguru import logger


@dataclass
class PairedSignificance:
    n: int
    "Number of paired samples"

    p_b_gt_t: float
    "Is target < baseline?"

    p_t_gt_b: float
    "Is target > baseline?"

    mu_B: float

    mu_T: float

    std: float

    @staticmethod
    def perform(baseline_values, target_values):
        B = np.array(baseline_values).flatten()
        T = np.array(target_values).flatten()
        assert (n := len(B)) == len(T)
        b_gt_t, t_gt_b = btest_rel(B, T)
        return PairedSignificance(
            n, b_gt_t, t_gt_b, B.mean(), T.mean(), (T - B).std(ddof=1)
        )

    def ratio(self):
        return max(self.p_b_gt_t / self.p_t_gt_b, self.p_t_gt_b / self.p_b_gt_t)


@dataclass
class GroupSignificance:
    n1: int
    n2: int
    # TODO: support


@dataclass
class Comparison:
    feature: Any
    baseline: RunScope
    target: RunScope
    sig: PairedSignificance


class ComparisonSet:
    def __init__(self, direction: Literal["min", "max"] = "min"):
        self.comparisons: list[Comparison] = []
        self.direction = direction

    def add(self, cmps: list[Comparison]):
        self.comparisons.extend(cmps)

    def table(self):
        from rich.table import Table
        from rich.text import Text
        from rich.pretty import Pretty

        table = Table(show_lines=True)
        table.add_column("Feature")
        table.add_column("Better")
        table.add_column(f"Comparison ({self.direction})")
        table.add_column("Worse")
        table.add_column("Significance", justify="right")
        table.add_column("n", justify="right")
        table.add_column("Δ", justify="right")
        table.add_column("μ", justify="right")
        table.add_column("σ", justify="right")

        self.comparisons.sort(key=lambda s: s.sig.ratio())
        for cmp in self.comparisons:
            sgn = 0  # index of "better" in [baseline, target]
            if cmp.sig.p_b_gt_t < cmp.sig.p_t_gt_b:  # assume min
                sgn = sgn ^ 1
            if self.direction == "max":
                sgn = sgn ^ 1

            scopes = [cmp.baseline, cmp.target]
            means = [cmp.sig.mu_B, cmp.sig.mu_T]
            better = sgn
            worse = sgn ^ 1

            table.add_row(
                cmp.feature,
                Pretty(scopes[better]),
                self.ratio_significance_description(cmp.sig.ratio()),
                Pretty(scopes[worse]),
                f"{cmp.sig.ratio():.03g}x",
                f"{cmp.sig.n}",
                f"{means[better] - means[worse]:.3g}",
                f"{means[better]:.3g}",
                f"{cmp.sig.std:.3g}",
            )

        return table

    def show_table(self):
        import rich

        rich.print(self.table())

    @staticmethod
    def ratio_significance_description(ratio: float):
        if ratio < 1:
            logger.error(f"Expected comparison ratio to be >= 1 (got {ratio})")

        if ratio < 2:
            return "[grey37]is similar to"
        elif ratio < 10:
            return "is better than"
        elif ratio < 100:
            return "[yellow]is much better than"
        else:
            return "[green]is overwhelmingly better than"


def grid_compare_frame(df: pd.DataFrame, baseline_sorter=lambda s: repr(s)):
    """
    Index set as scope columns.

    Count # samples in each grid.

    Row type design?
    """
    # baseline_sorter hack: repr => lexicographic order!

    metric_cols = [c for c in df.columns if c.startswith("metric:")]

    comparisons = []

    for metric in metric_cols:
        metric_df = df[[metric]].reset_index()
        scope_cols = list(metric_df.columns)
        scope_cols.remove(metric)

        def Not(*cols):
            return [c for c in metric_df.columns if c not in cols]

        for scope in scope_cols:
            scope_values = sorted(
                metric_df[scope].dropna().unique(), key=baseline_sorter
            )
            if len(scope_values) < 2:
                continue

            baseline_scope = scope_values[0]

            set_baseline = metric_df[metric_df[scope] == baseline_scope]

            for target_scope in scope_values[1:]:
                set_target = metric_df[metric_df[scope] == target_scope]

                set_combined = set_baseline.set_index(Not(metric, scope)).join(
                    set_target.set_index(Not(metric, scope)),
                    lsuffix="_baseline",
                    rsuffix="_target",
                    how="inner",
                )

                if (nsamples := len(set_combined)) < 3:
                    logger.warning(
                        f"Not enough samples ({nsamples}) to perform comparison of {baseline_scope} <-> {target_scope} for {metric}"
                    )
                    continue

                comparisons.append(
                    Comparison(
                        metric,
                        baseline_scope,
                        target_scope,
                        PairedSignificance.perform(
                            set_combined[f"{metric}_baseline"],
                            set_combined[f"{metric}_target"],
                        ),
                    )
                )
    return comparisons


def grid_compare(cqs: ChunkQuerySet, direction="min"):
    comparisons = []
    for af_type, df in cqs.metrics_view("scope").items():
        logger.info(f"Comparing {af_type}")
        comparisons.extend(grid_compare_frame(df))

    cset = ComparisonSet(direction=direction)
    cset.add(comparisons)
    return cset



from __future__ import annotations

from .scope_type import RunScope

from blake3 import blake3
import numpy as np

from typing import Optional


class Tagged:
    """
    A tag is an immutable set of scopes.
    At time of entry, a scope gets associated with an ID.
    IDs don't matter for equality and hash comparisons - they're mostly just for serialisation.
    Every logged object is associated with the current tag.
    """

    def __init__(self, scopes_in_context: dict[RunScope, str]):
        self.scope_state = {
            scope: context_id for scope, context_id in scopes_in_context.items()
        }  # copy

        # Validate: all records should belong to same namespace
        namespaces = set(
            scope.namespace()
            for scope in self.scope_state
            if not scope.namespace().startswith(".")
        )
        assert len(namespaces) <= 1, f"{namespaces}: Inconsistent namespaces"

        self.types = set(type(s) for s in self.scope_state.keys())

        if len(self.types) != len(self.scope_state):
            raise RuntimeError(
                f"Each scope time can only be entered once ({self.scope_state})"
            )

    def __repr__(self):
        scope_reprs = "\n".join("\t" + repr(s) for s in self)
        return f"Tagged {{\n{scope_reprs}\n}}"

    def __getitem__(self, scope_type: type[RunScope]):
        # TODO: validation
        for sc in self:
            if isinstance(sc, scope_type):
                return sc

    @staticmethod
    def empty():
        return Tagged({})

    def with_scope(self, scope, context_id: str) -> Tagged:
        assert scope not in self.scope_state
        return Tagged({**self.scope_state, scope: context_id})

    def without_scope(self, scope, context_id: Optional[str] = None) -> Tagged:
        # Can pass context_id for consistency check (enter/exit properly paired)
        if context_id is not None:
            assert self.scope_state[scope] == context_id

        sans = {**self.scope_state}
        del sans[scope]
        return Tagged(sans)

    def with_scopes(self, *more_scopes) -> Tagged:
        tag = self
        for s in more_scopes:
            tag = tag.with_scope(s)

    def without_scopes(self, *more_scopes) -> Tagged:
        raise RuntimeError("Deprecated!")
        assert set(more_scopes).issubset(self.scopes)
        return Tagged(self.scopes.difference(more_scopes))

    def matches_query(self, q: Tagged | list[RunScope] | set[RunScope]) -> bool:
        match q:
            case Tagged(scope_state=scope_state):
                return set(self.scope_state.keys()).issuperset(scope_state.keys())
            case (list() | set()) as scopes:
                return set(self.scope_state.keys()).issuperset(scopes)
            case _:
                raise TypeError(f"{q}")

    def hashed(self):
        sup_hash = np.frombuffer(
            b"0000000000000000000000000000000000000000000000000000000000000000",
            dtype=np.uint8,
        )
        for scope in self.scope_state.keys():
            # TODO: not proper hex
            hasher = blake3()
            hasher.update(scope.namespace().encode())
            hasher.update(type(scope).__name__.encode())
            hasher.update(scope.as_canonical_json().encode())
            sup_hash = np.maximum(
                sup_hash,
                np.frombuffer(hasher.hexdigest().encode(), dtype=np.uint8),
            )

        return sup_hash.tobytes().decode("utf-8")

    def __iter__(self):
        # Copy for safety to handle mutation during iteration
        yield from list(self.scope_state.keys())



from __future__ import annotations
from .scope_type import RunScope

import pyarrow as pa
import pyarrow.feather as ft
import pandas as pd
from blake3 import blake3
import attrs
import json
from io import BytesIO

from typing import Any, Self


class _CreateArtifact:
    def __init__(self, artifact_type):
        self.artifact_type = artifact_type
        self.registered = {}

    def __call__(self, *args, **kwargs):
        # To allow passing config for validation...
        return self

    def __getattr__(self, name):
        if name in self.registered:
            return self.registered[name]

        # TODO: support defining validators?
        new_artifact = attrs.frozen(
            type(
                name,
                (self.artifact_type,),
                {"__annotations__": {"data_hash": str}},
            )
        )
        self.registered[name] = new_artifact

        return new_artifact


class Artifact(RunScope):
    data_hash: str

    @classmethod
    def new(cls, obj) -> tuple[bytes, Self]:
        raise NotImplementedError()

    @staticmethod
    def type_string():
        raise NotImplementedError()


class JSONArtifact(Artifact):
    @classmethod
    def new(cls, obj) -> tuple[bytes, Self]:
        obj_bin = json.dumps(obj).encode()
        obj_hash = blake3().update(obj_bin).hexdigest()
        return obj_bin, cls(obj_hash)

    @classmethod
    def reconstruct(cls, data: bytes):
        return json.loads(data)

    def namespace(self):
        # special dot-prefixed namespace
        return ".Artifact.JSON"

    @staticmethod
    def type_string():
        return "JSON"


class TableArtifact(Artifact):
    @classmethod
    def new(cls, table) -> tuple[bytes, Self]:
        write_to = BytesIO()
        match table:
            case pd.DataFrame():
                table = pa.Table.from_pandas(table)
            case pa.Table():
                table = table
            case _:
                raise TypeError(f"{table}")

        with pa.ipc.new_file(write_to, table.schema) as writer:
            writer.write_table(table)

        obj_bin = write_to.getvalue()
        obj_hash = blake3().update(obj_bin).hexdigest()
        return obj_bin, cls(obj_hash)

    @classmethod
    def reconstruct(cls, data: bytes):
        return ft.read_table(BytesIO(data))

    def namespace(self):
        # special dot-prefixed namespace
        return ".Artifact.Table"

    @staticmethod
    def type_string():
        return "Table"


class Arty:
    JSON = _CreateArtifact(JSONArtifact)
    Table = _CreateArtifact(TableArtifact)

    @classmethod
    def any(cls, obj):
        """
        Unnamed default artifact scope.
        """
        match obj:
            case pd.DataFrame() | pa.Table():
                return cls.Table.AnyTable
            case dict():
                return cls.JSON.AnyJSON
            case _:
                raise TypeError(f"{type(obj)}")


"""
per[SavedMeasurements].load
for m in per[SavedMeasurements].query_all():
    with per.using(m) as df:
        computed_stuff = {"x": 100, "y": df["y"].sum()}
        # relax
        per.log(per.Arty.JSON.Computed, computed_stuff)

        # will autmatically log and upload?
        with per.context(Arty.JSON.Computed({"x": 1})):  # returns context
            Arty.JSON.Computed.load()  # will error if multiple found
            Arty.JSON.Computed.get()  # will

    # How does this handle using multiple of a type?
per.fetch_logs(SavedMeasurements)
"""



from permanence import ObjectPermanence, Arty
from permanence.diff import grid_compare

import pandas as pd
from random import random
import itertools


def test_grid_diff():
    pr = ObjectPermanence.new_namespace("test-diffs")

    @pr.new_scope
    class A(pr.Scope):
        shift: float

    @pr.new_scope
    class B(pr.Scope):
        dummy: str

    @pr.new_scope
    class C(pr.Scope):
        shift: float

    @pr.new_scope
    class X(pr.Scope):
        seed: int
        sigma: float

    def additive_model():
        point_sample = (random() - 1 / 2) * pr[X].sigma + pr[A].shift - pr[C].shift
        pr.logged(
            Arty.JSON.Additive,
            {"metric:score": point_sample, "metric:potato": random()},
        )  # Basic additive model

    # (A): +2.0
    # (B): +0.0 (no effect)
    # (C): -1.0
    # (X): random epsilon noise sampled some number of times
    param_space = (
        [A(0.0), A(2.0)],
        [B("placebo"), B("nocebo")],
        [C(0.0), C(1.0)],
        [X(s, 0.5) for s in range(5)],
    )
    # may want a Tag Exclusion filter (e.g. unwanted body parts)
    for ctx in itertools.product(*param_space):
        with pr.contexts(*ctx):
            additive_model()

    cmps = grid_compare(pr.query(), "max")
    cmps.show_table()

    # TODO: test sigs


if __name__ == "__main__":
    test_grid_diff()



import numpy as np
from scipy.stats import beta, gmean


def score_ranks(bounds):
    bounds = np.array(bounds)
    n = len(bounds)
    K = 1 + np.arange(n)
    X = np.sort(bounds)

    distribution = beta(K, n + 1 - K)
    cutoff = distribution.median()

    L = np.where(X > cutoff, 1, distribution.cdf(X) / distribution.cdf(cutoff))
    return gmean(L)


def bound_right(mean, median, stdev, x):
    a = x - mean
    k = a / stdev
    with np.errstate(divide="ignore"):
        # TODO: adjust stdev fudge using sample size
        return np.clip(
            np.minimum(
                np.where(x > median + stdev / 10, 0.5, 1.0),
                np.where(
                    a <= 0,
                    1.0,
                    np.where(
                        k >= np.sqrt(8 / 3), 4 / (9 * k**2), 4 / (3 * k**2) - 1 / 3
                    ),
                ),
                np.where(a <= 0, 1.0, stdev**2 / (stdev**2 + a**2)),
            ),
            0,
            1,
        )


def bound_left(mean, median, stdev, x):
    return bound_right(2 * x - mean, 2 * x - median, stdev, x)


def btest_rel(A, B):
    A = np.array(A)
    B = np.array(B)
    assert A.shape == B.shape and len(A.shape) == 1

    stdev = max(1e-8, np.std(A - B, ddof=1))

    p_a_gt_b = score_ranks(bound_right(0.0, 0.0, stdev, A - B))
    p_b_gt_a = score_ranks(bound_left(0.0, 0.0, stdev, A - B))
    return p_a_gt_b, p_b_gt_a



from .namespace import ScopeNamespace
from .scope_type import RunScope
from .tag_type import Tagged
from .artifact import Artifact, Arty
from .chunks import ChunkCollection, MutableLocalChunk, TaggedLocalArtifact

import time
from contextvars import ContextVar
from uuid6 import uuid7
from contextlib import contextmanager, ExitStack
from typing import Iterable, Callable, Any, TypeVar
from typing_extensions import dataclass_transform


class ObjectPermanence:
    """
    Stores objects, defined as one of:
    - Plain JSON object
    - Arrow table
    - Binary data

    You can choose to store pickled objects at your own risk.
    """

    def __init__(self, scope_namespace: ScopeNamespace):
        self.scope_namespace = scope_namespace
        self.namespace = scope_namespace.name

        self.current_tag = ContextVar("permanence_tag", default=Tagged.empty())
        self.current_chunk = MutableLocalChunk.new()
        # TODO: supports overlaying remote chunks
        self.chunks_overlay = ContextVar(
            "permanence_chunks_overlay", default=ChunkCollection([])
        )

    @staticmethod
    def new_namespace(name):
        return ObjectPermanence(ScopeNamespace(name))

    @dataclass_transform(frozen_default=True)
    def new_scope(self, cls):
        assert issubclass(cls, self.Scope)

        # disallow subclassing of other records
        scope_cls = [
            parent_cls for parent_cls in cls.mro() if issubclass(parent_cls, self.Scope)
        ]

        assert (
            len(scope_cls) <= 2
        ), f"{scope_cls}: Subclassing a Record type is disallowed"

        return self.scope_namespace.new_scope(cls)

    @property
    def Scope(self) -> type[RunScope]:
        return self.scope_namespace.Scope

    def active_scopes(self):
        return self.current_tag.get()

    @contextmanager
    def context(self, context: RunScope):
        # TODO: support multiple contexts
        assert isinstance(context, RunScope)

        # TODO: use contextvars reset token?
        entry_id = str(uuid7())
        self.current_tag.set(self.current_tag.get().with_scope(context, entry_id))

        try:
            yield context  # give back scope
        finally:
            self.current_tag.set(
                self.current_tag.get().without_scope(context, entry_id)
            )

    def contexts(self, *contexts: RunScope):
        # Immediately enters contexts, which could be problematic if ExitStack doesn't get entered
        stack = ExitStack()
        for c in contexts:
            stack.enter_context(self.context(c))
        return stack

    @contextmanager
    def using(self, af: Artifact | TaggedLocalArtifact):
        # TODO: support TaggedRemoteArtifact

        match af:
            case Artifact():
                scope = af
                # any instance will have identical content
                content = next(
                    self.chunks().fetch_instances(af)
                ).artifact_content  # Requires fetch
            case TaggedLocalArtifact():  # returned by the chunk
                scope = af.artifact
                content = af.artifact_content

        entry_id = str(uuid7())
        reset = self.current_tag.set(self.current_tag.get().with_scope(scope, entry_id))

        try:
            yield content
        finally:
            self.current_tag.reset(reset)

    @contextmanager
    def overlayed(self, overlay_chunks: ChunkCollection):
        reset_token = self.chunks_overlay.set(
            self.chunks_overlay.get() + overlay_chunks
        )
        try:
            yield
        finally:
            self.chunks_overlay.reset(reset_token)

    def chunks(self) -> ChunkCollection:
        # Shallow copy because we don't want interior mutability in ChunkCollection
        return self.chunks_overlay.get() + ChunkCollection([self.current_chunk.copy()])

    def query(self, *, scopes=(), values=None):
        return self.chunks().query([*self.active_scopes(), *scopes], values)

    def _log(self, artifact_type: type[Artifact], value_of) -> TaggedLocalArtifact:
        content_bin, value_scope = artifact_type.new(value_of)
        # TODO: write log time
        tla = TaggedLocalArtifact(
            value_of, value_scope, self.current_tag.get(), time.time()
        )
        self.current_chunk.append(content_bin, tla)
        return tla

    def log(self, value):
        """
        Untyped interface for logging arbitrary objects as artifacts.
        """
        artifact_type = Arty.any(value)
        return self.logged(artifact_type, value)

    def logged(self, artifact_type: type[Artifact], value_of):
        """
        Structured interface for logging typed and named artifacts.
        """
        tla = self._log(artifact_type, value_of)
        return tla.artifact

    def checkout(self, b3sum: str) -> Artifact:
        """
        To access `TaggedArtifact`s, use the query interface instead.
        """
        return next(self.chunks().fetch_instances_by_hash(b3sum)).artifact

    def try_checkout(self, b3sum: str):
        # TODO: -> Result[Artifact, None]
        raise NotImplementedError()

    T = TypeVar("T", bound=RunScope)

    def __getitem__(self, scope_or_artifact: type[T]) -> T:
        """
        Convenience interface to access the "current" scope

        where:
            S: RunScope excluding Artifact
        returns, given that:
        - type[S]: looks up and returns unique instance of that type from current scope (error if not found or non-unique)
        """
        if issubclass(scope_or_artifact, self.Scope):
            scopes = [
                s for s in self.active_scopes() if isinstance(s, scope_or_artifact)
            ]
            if len(scopes) != 1:
                raise RuntimeError(f"{scope_or_artifact} in {scopes}: Scope not active")
            return scopes[0]
        else:
            # Artifact or artifact type?
            raise NotImplementedError(scope_or_artifact)
